"""
Behavioral dynamics of component types.

@author: Gautham Ganapathy
@organization: LEMS (http://neuroml.org/lems/, https://github.com/organizations/LEMS)
@contact: gautham@lisphacker.org

MAvdVlag: altered attributes for state_variables, derived_variables, time_derivatives and
conditional_derived_variable.
"""

from ..base.base import LEMSBase
from ..base.map import Map
from ..base.errors import ModelError,ParseError

from ..parser.expr import ExprParser

class StateVariable(LEMSBase):
    """
    Store the specification of a state variable.
    """

    def __init__(self, name, default, boundaries = None):
        """
        Constructor.

        See instance variable documentation for more info on parameters.
        """

        self.name = name
        """ Name of the state variable.
        @type: str """

        self.default = default
        """ Dimension of the state variable.
        @type: str """

        self.boundaries = boundaries
        """ Exposure name for the state variable.
        @type: str """

    def __str__(self):
        return 'StateVariable name="{0}" dimension="{1}"'.format(self.name, self.default) +\
          (' exposure="{0}"'.format(self.boundaries) if self.boundaries else '')


    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<StateVariable name="{0}" dimension="{1}"'.format(self.name, self.dimension) +\
          (' exposure="{0}"'.format(self.exposure) if self.exposure else '') +\
          '/>'

class DerivedVariable(LEMSBase):
    """
    Store the specification of a derived variable.
    """

    def __init__(self, name, **params):
        """
        Constructor.

        See instance variable documentation for more info on parameters.
        """

        self.name = name
        """ Name of the derived variable.
        @type: str """

        self.dimension = params['dimension'] if 'dimension' in params else None
        """ Dimension of the derived variable or None if computed.
        @type: str """

        self.exposure = params['exposure'] if 'exposure' in params else None
        """ Exposure name for the derived variable.
        @type: str """

        self.select = params['select'] if 'select' in params else None
        """ Selection path/expression for the derived variable.
        @type: str """

        self.expression = params['expression'] if 'expression' in params else None
        """ Value of the derived variable.
        @type: str """

        self.reduce = params['reduce'] if 'reduce' in params else None
        """ Reduce method for the derived variable.
        @type: str """

        self.required = params['required'] if 'required' in params else None
        """ Requried or not.
        @type: str """

        self.expression_tree = None
        """ Parse tree for the time derivative expression.
        @type: lems.parser.expr.ExprNode """

        if self.expression != None:
            try:
                self.expression_tree = ExprParser(self.expression).parse()
            except:
                raise ParseError("Parse error when parsing value expression "
                                 "'{0}' for derived variable {1}",
                                 self.expression, self.name)

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<DerivedVariable name="{0}"'.format(self.name) +\
          (' dimension="{0}"'.format(self.dimension) if self.dimension else '') +\
          (' exposure="{0}"'.format(self.exposure) if self.exposure else '') +\
          (' select="{0}"'.format(self.select) if self.select else '') +\
          (' value="{0}"'.format(self.value) if self.value else '') +\
          (' reduce="{0}"'.format(self.reduce) if self.reduce else '') +\
          (' required="{0}"'.format(self.required) if self.required else '') +\
          '/>'
          
          
class Case(LEMSBase):
    """
    Store the specification of a case for a Conditional Derived Variable.
    """

    def __init__(self, condition, value):
        """
        Constructor.
        """

        self.condition = condition
        """ Condition for this case.
        @type: str """

        self.value = value
        """ Value if the condition is true.
        @type: str """

        self.condition_expression_tree = None
        """ Parse tree for the case condition expression.
        @type: lems.parser.expr.ExprNode """
        
        self.value_expression_tree = None
        """ Parse tree for the case condition expression.
        @type: lems.parser.expr.ExprNode """
        
        try:
            self.value_expression_tree = ExprParser(self.value).parse()
            
            if not self.condition:
                self.condition_expression_tree = None
            else:
                self.condition_expression_tree = ExprParser(self.condition).parse()
        except:
            raise ParseError("Parse error when parsing case with condition "
                             "'{0}' and value {1}",
                             self.condition, self.value)
        
    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<Case condition="{0}" value="{1}"'.format(self.condition, self.value) + '/>'
          
class ConditionalDerivedVariable(LEMSBase):
    """
    Store the specification of a conditional derived variable.
    """

    def __init__(self, name, condition, exposure = None, cases = None):
        """
        Constructor.

        See instance variable documentation for more info on parameters.
        """

        self.name = name
        """ Name of the derived variable.
        @type: str """

        self.condition = condition
        """ Dimension of the state variable.
        @type: str """

        self.exposure = exposure
        """ Exposure name for the state variable.
        @type: str """
        
        self.cases = list(cases.split(", "))
        """ List of cases related to this conditional derived variable.
        @type: list(lems.model.dynamics.Case) """
        
        
    def add_case(self, case):
        """
        Adds a case to this conditional derived variable.

        @param case: Case to be added.
        @type case: lems.model.dynamics.Case
        """

        self.cases.append(case)

    def add(self, child):
        """
        Adds a typed child object to the conditional derived variable.

        @param child: Child object to be added.
        """

        if isinstance(child, Case):
            self.add_case(child)
        else:
            raise ModelError('Unsupported child element')


    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        xmlstr = '<ConditionalDerivedVariable name="{0}"'.format(self.name) +\
          (' dimension="{0}"'.format(self.dimension) if self.dimension else '') +\
          (' exposure="{0}"'.format(self.exposure) if self.exposure else '') 
          
        chxmlstr = ''

        for case in self.cases:
            chxmlstr += case.toxml()

        if chxmlstr:
            xmlstr += '>' + chxmlstr + '</ConditionalDerivedVariable>'
        else:
            xmlstr += '/>'
            
        return xmlstr

class TimeDerivative(LEMSBase):
    """
    Store the specification of a time derivative specifcation.
    """

    def __init__(self, name, expression):
        """
        Constructor.

        See instance variable documentation for more info on parameters.
        """

        self.name = name
        """ Name of the variable for which the time derivative is being specified.
        @type: str """

        self.expression = expression
        """ Derivative expression.
        @type: str """

        self.expression_tree = None
        """ Parse tree for the time derivative expression.
        @type: lems.parser.expr.ExprNode """
        
        try:
            self.expression_tree = ExprParser(expression).parse()
        except:
            raise ParseError("Parse error when parsing value expression "
                             "'{0}' for state variable {1}",
                             self.expression, self.name)
        
    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<TimeDerivative variable="{0}" value="{1}"/>'.format(self.variable, self.value)

class Action(LEMSBase):
    """
    Base class for event handler actions.
    """

    pass

class StateAssignment(Action):
    """
    State assignment specification.
    """

    def __init__(self, variable, value):
        """
        Constructor.

        See instance variable documentation for more info on parameters.
        """

        Action.__init__(self)

        self.variable = variable
        """ Name of the variable for which the time derivative is being specified.
        @type: str """

        self.value = value
        """ Derivative expression.
        @type: str """

        self.expression_tree = None
        """ Parse tree for the time derivative expression.
        @type: lems.parser.expr.ExprNode """

        try:
            self.expression_tree = ExprParser(value).parse()
        except:
            raise ParseError("Parse error when parsing state assignment "
                             "value expression "
                             "'{0}' for state variable {1}",
                             self.value, self.variable)

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<StateAssignment variable="{0}" value="{1}"/>'.format(self.variable, self.value)


class EventOut(Action):
    """
    Event transmission specification.
    """

    def __init__(self, port):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """
        
        Action.__init__(self)

        self.port = port
        """ Port on which the event comes in.
        @type: str """
        
    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<EventOut port="{0}"/>'.format(self.port)

class Transition(Action):
    """
    Regime transition specification.
    """

    def __init__(self, regime):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """
        
        Action.__init__(self)

        self.regime = regime
        """ Regime to transition to.
        @type: str """

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return '<Transition regime="{0}"/>'.format(self.regime)

class EventHandler(LEMSBase):
    """
    Base class for event handlers.
    """

    def __init__(self):
        """
        Constructor.
        """

        self.actions = list()
        """ List of actions to be performed in response to this event.
        @type: list(lems.model.dynamics.Action) """

    def __str__(self):
        istr = 'EventHandler...'
        return istr

    def add_action(self, action):
        """
        Adds an action to this event handler.

        @param action: Action to be added.
        @type action: lems.model.dynamics.Action
        """

        self.actions.append(action)

    def add(self, child):
        """
        Adds a typed child object to the event handler.

        @param child: Child object to be added.
        """

        if isinstance(child, Action):
            self.add_action(child)
        else:
            raise ModelError('Unsupported child element')
        
class OnStart(EventHandler):
    """
    Specification for event handler called upon initialization of the component.
    """

    def __init__(self):
        """
        Constructor.
        """
        
        EventHandler.__init__(self)

    def __str__(self):
        istr = 'OnStart: ['
        for action in self.actions:
            istr += str(action)
        istr += ']'
        return str(istr)

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        xmlstr = '<OnStart'

        chxmlstr = ''

        for action in self.actions:
            chxmlstr += action.toxml()

        if chxmlstr:
            xmlstr += '>' + chxmlstr + '</OnStart>'
        else:
            xmlstr += '/>'

        return xmlstr

class OnCondition(EventHandler):
    """
    Specification for event handler called upon satisfying a given condition.
    """

    def __init__(self, test):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """
        
        EventHandler.__init__(self)

        self.test = test
        """ Condition to be tested for.
        @type: str """

        try:
            self.expression_tree = ExprParser(test).parse()
        except:
            raise ParseError("Parse error when parsing OnCondition test '{0}'", test)
        
    def __str__(self):
        istr = 'OnCondition...'
        return istr

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        xmlstr = '<OnCondition test="{0}"'.format(self.test)

        chxmlstr = ''

        for action in self.actions:
            chxmlstr += action.toxml()

        if chxmlstr:
            xmlstr += '>' + chxmlstr + '</OnCondition>'
        else:
            xmlstr += '/>'

        return xmlstr

class OnEvent(EventHandler):
    """
    Specification for event handler called upon receiving en event sent by another component.
    """

    def __init__(self, port):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """
        
        EventHandler.__init__(self)

        self.port = port
        """ Port on which the event comes in.
        @type: str """

    def __str__(self):
        istr = 'OnEvent, port: %s'%self.port
        return istr

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        xmlstr = '<OnEvent port="{0}"'.format(self.port)

        chxmlstr = ''

        for action in self.actions:
            chxmlstr += action.toxml()

        if chxmlstr:
            xmlstr += '>' + chxmlstr + '</OnEvent>'
        else:
            xmlstr += '/>'

        return xmlstr

class OnEntry(EventHandler):
    """
    Specification for event handler called upon entry into a new behavior regime.
    """

    def __init__(self):
        """
        Constructor.
        """
        
        EventHandler.__init__(self)

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        xmlstr = '<OnEntry'

        chxmlstr = ''

        for action in self.actions:
            chxmlstr += action.toxml()

        if chxmlstr:
            xmlstr += '>' + chxmlstr + '</OnEntry>'
        else:
            xmlstr += '/>'

        return xmlstr

class KineticScheme(LEMSBase):
    """
    Kinetic scheme specifications.
    """

    def __init__(self, name, nodes, state_variable, 
                 edges, edge_source, edge_target,
                 forward_rate, reverse_rate):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """

        self.name = name
        """ Name of the kinetic scheme.
        @type: str """

        self.nodes = nodes
        """ Nodes to be used for the kinetic scheme.
        @type: str """

        self.state_variable = state_variable
        """ State variable updated by the kinetic scheme.
        @type: str """

        self.edges = edges
        """ Edges to be used for the kinetic scheme.
        @type: str """

        self.edge_source = edge_source
        """ Attribute that defines the source of the transition.
        @type: str """

        self.edge_target = edge_target
        """ Attribute that defines the target of the transition.
        @type: str """

        self.forward_rate = forward_rate
        """ Name of the forward rate exposure.
        @type: str """

        self.reverse_rate = reverse_rate
        """ Name of the reverse rate exposure.
        @type: str """

    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        return ('<KineticScheme '
                'name="{0}" '
                'nodes="{1}" '
                'edges="{2}" '
                'stateVariable="{3}" '
                'edgeSource="{4}" '
                'edgeTarget="{5}" '
                'forwardRate="{6}" '
                'reverseRate="{7}"/>').format(self.name,
                                              self.nodes,
                                              self.edges,
                                              self.state_variable,
                                              self.edge_source,
                                              self.edge_target,
                                              self.forward_rate,
                                              self.reverse_rate)

class Behavioral(LEMSBase):
    """
    Store dynamic behavioral attributes.
    """

    def __init__(self):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """

        self.parent_behavioral = None
        """ Parent behavioral object.
        @type: lems.model.dynamics.Behavioral """

        self.state_variables = Map()
        """ Map of state variables in this behavior regime.
        @type: dict(str -> lems.model.dynamics.StateVariable """

        self.derived_variables = Map()
        """ Map of derived variables in this behavior regime.
        @type: dict(str -> lems.model.dynamics.DerivedVariable """

        self.conditional_derived_variables = Map()
        """ Map of conditional derived variables in this behavior regime.
        @type: dict(str -> lems.model.dynamics.ConditionalDerivedVariable """

        self.time_derivatives = Map()
        """ Map of time derivatives in this behavior regime.
        @type: dict(str -> lems.model.dynamics.TimeDerivative) """

        self.event_handlers = list()
        """ List of event handlers in this behavior regime.
        @type: list(lems.model.dynamics.EventHandler) """

        self.kinetic_schemes = Map()
        """ Map of kinetic schemes in this behavior regime.
        @type: dict(str -> lems.model.dynamics.KineticScheme) """
        
        
    def has_content(self):
        if len(self.state_variables)==0 and \
           len(self.derived_variables)==0 and \
           len(self.conditional_derived_variables)==0 and \
           len(self.time_derivatives)==0 and \
           len(self.event_handlers)==0 and \
           len(self.kinetic_schemes)==0:
               return False
        else:
            return True
        
    def clear(self):
        """
        Clear behavioral entities.
        """
        
        self.time_derivatives = Map()

    def add_state_variable(self, sv):
        """
        Adds a state variable to this behavior regime.

        @param sv: State variable.
        @type sv: lems.model.dynamics.StateVariable
        """

        self.state_variables[sv.name] = sv
        
    def add_derived_variable(self, dv):
        """
        Adds a derived variable to this behavior regime.

        @param dv: Derived variable.
        @type dv: lems.model.dynamics.DerivedVariable
        """

        self.derived_variables[dv.name] = dv
        
    def add_conditional_derived_variable(self, cdv):
        """
        Adds a conditional derived variable to this behavior regime.

        @param cdv: Conditional Derived variable.
        @type cdv: lems.model.dynamics.ConditionalDerivedVariable
        """

        self.conditional_derived_variables[cdv.name] = cdv
        
    def add_time_derivative(self, td):
        """
        Adds a time derivative to this behavior regime.

        @param td: Time derivative.
        @type td: lems.model.dynamics.TimeDerivative
        """

        self.time_derivatives[td.name] = td

    def add_event_handler(self, eh):
        """
        Adds an event handler to this behavior regime.

        @param eh: Event handler.
        @type eh: lems.model.dynamics.EventHandler
        """
        
        self.event_handlers.append(eh)

    def add_kinetic_scheme(self, ks):
        """
        Adds a kinetic scheme to this behavior regime.

        @param ks: Kinetic scheme.
        @type ks: lems.model.dynamics.KineticScheme
        """

        self.kinetic_schemes[ks.name] = ks

    def add(self, child):
        """
        Adds a typed child object to the behavioral object.

        @param child: Child object to be added.
        """

        if isinstance(child, StateVariable):
            self.add_state_variable(child)
        elif isinstance(child, DerivedVariable):
            self.add_derived_variable(child)
        elif isinstance(child, ConditionalDerivedVariable):
            self.add_conditional_derived_variable(child)
        elif isinstance(child, TimeDerivative):
            self.add_time_derivative(child)
        elif isinstance(child, EventHandler):
            self.add_event_handler(child)
        elif isinstance(child, KineticScheme):
            self.add_kinetic_scheme(child)
        else:
            raise ModelError('Unsupported child element')
        
    def toxml(self):
        """
        Exports this object into a LEMS XML object
        """

        chxmlstr = ''

        for state_variable in self.state_variables:
            chxmlstr += state_variable.toxml()

        for derived_variable in self.derived_variables:
            chxmlstr += derived_variable.toxml()
            
        for conditional_derived_variable in self.conditional_derived_variables:
            chxmlstr += conditional_derived_variable.toxml()

        for time_derivative in self.time_derivatives:
            chxmlstr += time_derivative.toxml()

        for event_handler in self.event_handlers:
            chxmlstr += event_handler.toxml()

        for kinetic_scheme in self.kinetic_schemes:
            chxmlstr += kinetic_scheme.toxml()

        if isinstance(self, Dynamics):
            for regime in self.regimes:
                chxmlstr += regime.toxml()

        if isinstance(self, Dynamics):
            xmlprefix = 'Dynamics'
            xmlsuffix = 'Dynamics'
            xmlempty = ''
        else:
            xmlprefix = 'Regime name="{0}"'.format(self.name) +\
              (' initial="true"' if self.initial else '')
            xmlsuffix = 'Regime'
            xmlempty = '<{0}/>',format(xmlprefix)

        if chxmlstr:
            xmlstr = '<{0}>'.format(xmlprefix) + chxmlstr + '</{0}>'.format(xmlsuffix)
        else:
            xmlstr = xmlempty

        return xmlstr
                
class Regime(Behavioral):
    """
    Stores a single behavioral regime for a component type.
    """

    def __init__(self, name, parent_behavioral, initial = False):
        """
        Constructor.
        
        See instance variable documentation for more details on parameters.
        """

        Behavioral.__init__(self)
        
        self.name = name
        """ Name of this behavior regime.
        @type: str """

        self.parent_behavioral = parent_behavioral
        """ Parent behavioral object.
        @type: lems.model.dynamics.Behavioral """

        self.initial = initial
        """ Initial behavior regime.
        @type: bool """
        
class Dynamics(Behavioral):
    """
    Stores behavioral dynamics specification for a component type.
    """

    def __init__(self):
        """
        Constructor.
        """
        
        Behavioral.__init__(self)

        self.regimes = Map()
        """ Map of behavior regimes.
        @type: Map(str -> lems.model.dynamics.Regime) """

    def add_regime(self, regime):
        """
        Adds a behavior regime to this dynamics object.

        @param regime: Behavior regime to be added.
        @type regime: lems.model.dynamics.Regime """

        self.regimes[regime.name] = regime

    def add(self, child):
        """
        Adds a typed child object to the dynamics object.

        @param child: Child object to be added.
        """

        if isinstance(child, Regime):
            self.add_regime(child)
        else:
            Behavioral.add(self, child)
            
    def has_content(self):
        if len(self.regimes)>0: 
            return True
        else:
            return Behavioral.has_content(self)
        



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Perform Principal Component Analysis (PCA) on a TimeSeries datatype and return
a PrincipalComponents datatype.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import numpy
import tvb.datatypes.time_series as time_series
import tvb.datatypes.mode_decompositions as mode_decompositions
from tvb.basic.neotraits.api import HasTraits, Attr, narray_describe


class PCA_mlab(object):
    """
    The code for this method has been taken and adapted from Matplotlib 2.1.0
    Aug 2019
    """

    def __init__(self, data):
        n, m = data.shape
        if n < m:
            raise RuntimeError('we assume data in input is organized with '
                               'numrows>numcols')

        self.mu = data.mean(axis=0)
        self.sigma = data.std(axis=0)

        a = (data - self.mu) / self.sigma

        U, s, Vh = numpy.linalg.svd(a, full_matrices=False)

        self.Wt = Vh

        Y = numpy.dot(Vh, a.T).T
        self.Y = Y

        self.s = s ** 2

        vars = self.s / len(s)
        self.fracs = vars / vars.sum()


class PCA(HasTraits):
    """
    Return principal component weights and the fraction of the variance that 
    they explain. 
    
    PCA takes time-points as observations and nodes as variables.
    
    NOTE: The TimeSeries must be longer(more time-points) than the number of
          nodes -- Mostly a problem for TimeSeriesSurface datatypes, which, if 
          sampled at 1024Hz, would need to be greater than 16 seconds long.
    """

    time_series = Attr(
        field_type=time_series.TimeSeries,
        label="Time Series",
        required=True,
        doc="""The timeseries to which the PCA is to be applied. NOTE: The 
            TimeSeries must be longer(more time-points) than the number of nodes
            -- Mostly a problem for surface times-series, which, if sampled at
            1024Hz, would need to be greater than 16 seconds long.""")

    # TODO: Maybe should support first N components or neccessary components to
    #      explain X% of the variance. NOTE: For default surface the weights
    #      matrix has a size ~ 2GB * modes * vars...

    def evaluate(self):
        """
        Compute the temporal covariance between nodes in the time_series. 
        """
        cls_attr_name = self.__class__.__name__ + ".time_series"
        # self.time_series.trait["data"].log_debug(owner = cls_attr_name)

        ts_shape = self.time_series.data.shape

        # Need more measurements than variables
        if ts_shape[0] < ts_shape[2]:
            msg = "PCA requires a longer timeseries (tpts > number of nodes)."
            self.log.error(msg)
            raise Exception(msg)

        # (nodes, nodes, state-variables, modes)
        weights_shape = (ts_shape[2], ts_shape[2], ts_shape[1], ts_shape[3])
        self.log.info("weights shape will be: %s" % str(weights_shape))

        fractions_shape = (ts_shape[2], ts_shape[1], ts_shape[3])
        self.log.info("fractions shape will be: %s" % str(fractions_shape))

        weights = numpy.zeros(weights_shape)
        fractions = numpy.zeros(fractions_shape)

        # One inter-node temporal covariance matrix for each state-var & mode.
        for mode in range(ts_shape[3]):
            for var in range(ts_shape[1]):
                data = self.time_series.data[:, var, :, mode]

                data_pca = PCA_mlab(data)
                fractions[:, var, mode] = data_pca.fracs

                weights[:, :, var, mode] = data_pca.Wt

        self.log.debug("fractions")
        self.log.debug(narray_describe(fractions))
        self.log.debug("weights")
        self.log.debug(narray_describe(weights))

        pca_result = mode_decompositions.PrincipalComponents(
            source=self.time_series,
            fractions=fractions,
            weights=weights,
            norm_source=numpy.array([]),
            component_time_series=numpy.array([]),
            normalised_component_time_series=numpy.array([]))

        return pca_result

    def result_shape(self, input_shape):
        """
        Returns the shape of the main result of the PCA analysis -- compnnent 
        weights matrix and a vector of fractions.
        """
        weights_shape = (input_shape[2], input_shape[2], input_shape[1],
                         input_shape[3])
        fractions_shape = (input_shape[2], input_shape[1], input_shape[3])
        return [weights_shape, fractions_shape]

    def result_size(self, input_shape):
        """
        Returns the storage size in Bytes of the results of the PCA analysis.
        """
        result_size = numpy.sum(list(map(numpy.prod,
                                         self.result_shape(input_shape)))) * 8.0  # Bytes
        return result_size

    def extended_result_size(self, input_shape):
        """
        Returns the storage size in Bytes of the extended result of the PCA.
        That is, it includes storage of the evaluated PrincipleComponents
        attributes such as norm_source, component_time_series, etc.
        """
        result_size = self.result_size(input_shape)
        extend_size = result_size  # Main arrays
        extend_size = extend_size + numpy.prod(input_shape) * 8.0  # norm_source
        extend_size = extend_size + numpy.prod(input_shape) * 8.0  # component_time_series
        extend_size = extend_size + numpy.prod(input_shape) * 8.0  # normalised_component_time_series
        return extend_size



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details. You should have received a copy of the GNU
#  General Public License along with this program.  If not,
# see <http://www.gnu.org/licenses/>.
#
#
# CITATION: When using The Virtual Brain for scientific publications, please
# cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)

"""
Saggio codimension 3 Epileptor model

.. moduleauthor:: Len Spek

"""

import numpy

from .base import ModelNumbaDfun
from numba import guvectorize, float64, int64
from tvb.basic.neotraits.api import NArray, List, Range, Final


class EpileptorCodim3(ModelNumbaDfun):
    r"""
    .. [Saggioetal_2017] Saggio ML, Spiegler A, Bernard C, Jirsa VK.
    *Fast–Slow Bursters in the Unfolding of a High Codimension Singularity
    and the Ultra-slow Transitions of Classes.* Journal of Mathematical
    Neuroscience. 2017;7:7. doi:10.1186/s13408-017-0050-8.

    .. The Epileptor codim 3 model is a neural mass model which contains two
    subsystems acting at different timescales. For the fast subsystem we use
    the unfolding of a degenerate Takens-Bogdanov bifucation of codimension
    3. The slow subsystem steers the fast one back and forth along paths
    leading to bursting behavior. The model is able to produce almost all the
    classes of bursting predicted for systems with a planar fast subsystem.

    .. In this implementation the model can produce Hysteresis-Loop bursters
    of classes c0, c0', c2s, c3s, c4s, c10s, c11s, c2b, c4b, c8b, c14b and
    c16b as classified by [Saggioetal_2017] Table 2. The default model
    parameters correspond to class c2s.

    """

    mu1_start = NArray(
        label=":math:`mu_1 start`",
        default=numpy.array([-0.02285]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu1 at the offset point for the given class, default for class c2s "
            "(Saddle-Node at onset and Saddle-Homoclinic at offset)")

    mu2_start = NArray(
        label=":math:`mu_2 start`",
        default=numpy.array([0.3448]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation mu2 parameter at the offset point for the given class, default for class c2s "
            "(Saddle-Node at onset and Saddle-Homoclinic at offset)")

    nu_start = NArray(
        label=":math:`nu start`",
        default=numpy.array([0.2014]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation nu parameter at the offset point for the given class, default for class c2s "
            "(Saddle-Node at onset and Saddle-Homoclinic at offset)")

    mu1_stop = NArray(
        label=":math:`mu_1 stop`",
        default=numpy.array([-0.07465]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation mu1 parameter at the onset point for the given class, default for class c2s "
            "(Saddle-Node at onset and Saddle-Homoclinic at offset)")

    mu2_stop = NArray(
        label=":math:`mu_2 stop`",
        default=numpy.array([0.3351]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation mu2 parameter at the onset point for the given class, default for class c2s "
            "(Saddle-Node at onset and Saddle-Homoclinic at offset)")

    nu_stop = NArray(
        label=":math:`nu stop`",
        default=numpy.array([0.2053]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation nu parameter at the onset point for the given class, default for class c2s "
            "(Saddle-Node at onset and Saddle-Homoclinic at offset)")

    b = NArray(
        label=":math:`b`",
        default=numpy.array([1.0]),
        doc="Unfolding type of the degenerate Takens-Bogdanov bifurcation, default is a focus type")

    R = NArray(
        label=":math:`R`",
        default=numpy.array([0.4]),
        domain=Range(lo=0.0, hi=2.5),
        doc="Radius in unfolding")

    c = NArray(
        label=":math:`c`",
        default=numpy.array([0.001]),
        domain=Range(lo=0.0, hi=0.01),
        doc="Speed of the slow variable")

    dstar = NArray(
        label=":math:`d^*`",
        default=numpy.array([0.3]),
        domain=Range(lo=-0.1, hi=0.5),
        doc="Threshold for the inversion of the slow variable")

    Ks = NArray(
        label=":math:`K_s`",
        default=numpy.array([0.0]),
        doc="Slow permittivity coupling strength, the default is no coupling")

    N = NArray(
        dtype=int,
        label=":math:`N`",
        default=numpy.array([1]),
        doc="The branch of the resting state, default is 1")

    modification = NArray(
        dtype=bool,
        label=":math:`modification`",
        default=numpy.array([True]),
        doc="When modification is True, then use the modification to stabilise the system for negative values of "
            "dstar. If modification is False, then don't use the modification. The default value is True ")

    state_variable_range = Final(
        label="State variable ranges [lo, hi]",
        default={"x": numpy.array([0.4, 0.6]),
                 "y": numpy.array([-0.1, 0.1]),
                 "z": numpy.array([0.0, 0.15])},
        doc="Typical bounds on state variables.")

    variables_of_interest = List(
        of=str,
        label="Variables watched by Monitors",
        choices=('x', 'y', 'z'),
        default=('x', 'z'),
        doc="Quantities available to monitor.")

    # state variables names
    state_variables = ('x', 'y', 'z')

    # number of state variables
    _nvar = 3
    cvar = numpy.array([0], dtype=numpy.int32)

    # If there are derived parameters from the predefined parameters, then initialize them to None
    G = None
    H = None
    L = None
    M = None

    def update_derived_parameters(self):
        r"""
        The equations were taken from [Saggioetal_2017]
        cf. Eqn. (7), page 17

        Here we parametrize the great arc which lies on a sphere of radius R
        between the points A and B, which are given by:

            .. math::
                A &= \begin{pmatrix}\mu_{2,start} & -\mu_{1,start} & \nu_{start} \end{pmatrix} \\
                B &= \begin{pmatrix}\mu_{2,stop} & -\mu_{1,stop} & \nu_{stop} \end{pmatrix}

        Then we parametrize this great arc with z as parameter by :math:`R(E \cos z + F \sin z)`
            where the unit vectors E and F are given by:

            .. math::
                E &= A/\|A\| \\
                F &= ((A \times B) \times A)/\|(A \times B) \times A\|
        """

        A = numpy.array(
            [self.mu2_start[0], -self.mu1_start[0], self.nu_start[0]])
        B = numpy.array([self.mu2_stop[0], -self.mu1_stop[0], self.nu_stop[0]])

        self.E = A / numpy.linalg.norm(A)
        self.F = numpy.cross(numpy.cross(A, B), A)
        self.F = self.F / numpy.linalg.norm(self.F)

    def _numpy_dfun(self, state_variables, coupling, local_coupling=0.0):
        x = state_variables[0, :]
        y = state_variables[1, :]
        z = state_variables[2, :]

        # Computes the values of mu2,mu1 and nu given the great arc (E,F,R) and the value of the slow variable z
        mu2 = self.R * (self.E[0] * numpy.cos(z) + self.F[0] * numpy.sin(z))
        mu1 = -self.R * (self.E[1] * numpy.cos(z) + self.F[1] * numpy.sin(z))
        nu = self.R * (self.E[2] * numpy.cos(z) + self.F[2] * numpy.sin(z))

        # Computes x_s, which is the solution to x_s^3 - mu2*x_s - mu1 = 0
        if self.N == 1:
            xs = (mu1 / 2.0 + numpy.sqrt(
                mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) + (mu1 / 2.0 - numpy.sqrt(
                mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0)
        elif self.N == 2:
            xs = -1.0 / 2.0 * (1.0 - 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 + numpy.sqrt(
                mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) - 1.0 / 2.0 * (
                         1.0 + 1j * 3 ** (1.0 / 2.0)) * (
                             mu1 / 2.0 - numpy.sqrt(mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (
                         1.0 / 3.0)
        elif self.N == 3:
            xs = -1.0 / 2.0 * (1.0 + 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 + numpy.sqrt(
                mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) - 1.0 / 2.0 * (
                         1.0 - 1j * 3 ** (1.0 / 2.0)) * (
                             mu1 / 2.0 - numpy.sqrt(mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (
                         1.0 / 3.0)
        xs = numpy.real(xs)

        xdot = -y
        ydot = x ** 3 - mu2 * x - mu1 - y * (nu + self.b * x + x ** 2)
        if self.modification:
            zdot = -self.c * (
                    numpy.sqrt((x - xs) ** 2 + y ** 2) - self.dstar + 0.1 * (z - 0.5) ** 7 + self.Ks * coupling[0, :])
        else:
            zdot = -self.c * (numpy.sqrt(
                (x - xs) ** 2 + y ** 2) - self.dstar + self.Ks * coupling[0, :])

        derivative = numpy.array([xdot, ydot, zdot])
        return derivative

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
        The equations were taken from [Saggioetal_2017]
        cf. Eqns. (4) and (7), page 17

        The state variables :math:`x` and :math:`y` correspond to the fast subsystem and the
        state variable :math:`z` corresponds to the slow subsystem.

        .. math::
                \dot{x} &= -y \\
                \dot{y} &= x^3 - \mu_{2} x - \mu_{1} - y(\nu + b x + x^2) \\
                \dot{z} &= -c(\sqrt{x-x_{s}^2+y^2} - d^*)

        If the bool modification is True, then the equation for :math:`\dot{z}` will
        been modified to ensure stability for negative :math:`d^x`

        .. math::
            \dot{z} = -c(\sqrt{x-x_{s}^2+y^2} - d^* + 0.1(z-0.5)^7)

        Where :math:`\mu_1, \mu_2` and :math:`\nu` lie on a great arc of a
        sphere of radius :math:`R` parametrised by the unit vectors :math:`E` and :math:`F`.

        .. math::
            \begin{pmatrix}\mu_2 & -\mu_1 & \nu \end{pmatrix} = R(E \cos z + F \sin z)

        And where :math:`x_s` is the x-coordinate of the resting state
        (stable equilibrium). This is computed by finding the solution of

            .. math::
                x_s^3 - mu_2*x_s - mu_1 = 0

        And taking the branch which corresponds to the resting state.
        If :math:`x_s` is complex, we take the real part.
        """
        state_variables_ = state_variables.reshape(state_variables.shape[:-1]).T
        coupling_ = coupling.reshape(coupling.shape[:-1]).T
        derivative = _numba_dfun(state_variables_, coupling_, self.E[0], self.E[1], self.E[2], self.F[0], self.F[1],
                                 self.F[2], self.b, self.R, self.c, self.dstar, self.Ks, self.modification, self.N)
        return derivative.T[..., numpy.newaxis]


@guvectorize([(float64[:], float64[:], float64[:], float64[:], float64[:], float64[:], float64[:], float64[:],
               float64[:], float64[:], float64[:], float64[:], float64[:], int64[:], int64[:], float64[:])],
             '(n),(m)' + ',()' * 13 + '->(n)', nopython=True)
def _numba_dfun(state_variables, coupling, E0, E1, E2, F0, F1, F2, b, R, c, dstar, Ks, modification, N, derivative):
    """Gufunction for the Epileptor Codim 3 model"""

    x = state_variables[0]
    y = state_variables[1]
    z = state_variables[2]

    # Computes the values of mu2,mu1 and nu given the great arc (E,F,R) and the value of the slow variable z
    mu2 = R[0] * (E0[0] * numpy.cos(z) + F0[0] * numpy.sin(z))
    mu1 = -R[0] * (E1[0] * numpy.cos(z) + F1[0] * numpy.sin(z))
    nu = R[0] * (E2[0] * numpy.cos(z) + F2[0] * numpy.sin(z))

    # Computes x_s, which is the solution to x_s^3 - mu2*x_s - mu1 = 0
    if N[0] == 1:
        xs = (mu1 / 2.0 + numpy.sqrt(
            mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) + (mu1 / 2.0 - numpy.sqrt(
            mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0)
    elif N[0] == 2:
        xs = -1.0 / 2.0 * (1.0 - 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 + numpy.sqrt(
            mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) - 1.0 / 2.0 * (
            1.0 + 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 - numpy.sqrt(mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (
            1.0 / 3.0)
    elif N[0] == 3:
        xs = -1.0 / 2.0 * (1.0 + 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 + numpy.sqrt(
            mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) - 1.0 / 2.0 * (
            1.0 - 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 - numpy.sqrt(mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (
            1.0 / 3.0)
    xs = xs.real

    derivative[0] = -y
    derivative[1] = x ** 3 - mu2 * x - mu1 - y * (nu + b[0] * x + x ** 2)
    derivative[2] = -c[0] * (
        numpy.sqrt((x - xs) ** 2 + y ** 2) - dstar[0] + modification[0] * 0.1 * (z - 0.5) ** 7 + Ks[0] * coupling[0])


class EpileptorCodim3SlowMod(ModelNumbaDfun):
    r"""
    .. [Saggioetal_2017] Saggio ML, Spiegler A, Bernard C, Jirsa VK.
    *Fast–Slow Bursters in the Unfolding of a High Codimension Singularity
    and the Ultra-slow Transitions of Classes.* Journal of Mathematical
    Neuroscience. 2017;7:7. doi:10.1186/s13408-017-0050-8.

    .. The Epileptor codim 3 model is a neural mass model which contains two
    subsystems acting at different timescales. For the fast subsystem we use
    the unfolding of a degenerate Takens-Bogdanov bifucation of codimension
    3. The slow subsystem steers the fast one back and forth along these
    paths leading to bursting behavior. The model is able to produce almost
    all the classes of bursting predicted for systems with a planar fast
    subsystem.

    .. In this implementation the model can produce Hysteresis-Loop bursters
    of classes c0, c0', c2s, c3s, c4s, c10s, c11s, c2b, c4b, c8b, c14b and
    c16b as classified by [Saggioetal_2017] Table 2. Through ultra-slow
    modulation of the path through the parameter space we can switch between
    different classes of bursters.

    """

    mu1_Ain = NArray(
        label="mu1 Ain",
        default=numpy.array([0.05494]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu1 at the initial point at bursting offset.")

    mu2_Ain = NArray(
        label="mu2 Ain",
        default=numpy.array([0.2731]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu2 at the initial point at bursting offset.")

    nu_Ain = NArray(
        label="nu Ain",
        default=numpy.array([0.287]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter nu at the initial point at bursting offset.")

    mu1_Bin = NArray(
        label="mu1 Bin",
        default=numpy.array([-0.0461]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu1 at the initial point at bursting onset.")

    mu2_Bin = NArray(
        label="mu2 Bin",
        default=numpy.array([0.243]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu2 at the initial point at bursting onset.")

    nu_Bin = NArray(
        label="nu Bin",
        default=numpy.array([0.3144]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter nu at the initial point at bursting onset.")

    mu1_Aend = NArray(
        label="mu1 Aend",
        default=numpy.array([0.06485]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu1 at the initial point at bursting offset.")

    mu2_Aend = NArray(
        label="mu2 Aend",
        default=numpy.array([0.07337]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu2 at the initial point at bursting offset.")

    nu_Aend = NArray(
        label="nu Aend",
        default=numpy.array([-0.3878]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter nu at the initial point at bursting offset.")

    mu1_Bend = NArray(
        label="mu1 Bend",
        default=numpy.array([0.03676]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu1 at the initial point at bursting onset.")

    mu2_Bend = NArray(
        label="mu2 Bend",
        default=numpy.array([-0.02792]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter mu2 at the initial point at bursting onset.")

    nu_Bend = NArray(
        label="nu Bend",
        default=numpy.array([-0.3973]),
        domain=Range(lo=-1.0, hi=1.0),
        doc="The bifurcation parameter nu at the initial point at bursting onset.")

    b = NArray(
        label="b",
        default=numpy.array([1.0]),
        doc="Unfolding type of the degenerate Takens-Bogdanov bifurcation, default is a focus type")

    R = NArray(
        label="R",
        default=numpy.array([0.4]),
        domain=Range(lo=0.0, hi=2.5),
        doc="Radius in unfolding")

    c = NArray(
        label="c",
        default=numpy.array([0.002]),
        domain=Range(lo=0.0, hi=0.01),
        doc="Speed of the slow variable")

    cA = NArray(
        label="cA",
        default=numpy.array([0.0001]),
        domain=Range(lo=0.0, hi=0.001),
        doc="Speed of the ultra-slow transition of the initial point")

    cB = NArray(
        label="cB",
        default=numpy.array([0.00012]),
        domain=Range(lo=0.0, hi=0.001),
        doc="Speed of the ultra-slow transition of the final point")

    dstar = NArray(
        label="dstar",
        default=numpy.array([0.3]),
        domain=Range(lo=-0.1, hi=0.5),
        doc="Threshold for the inversion of the slow variable")

    Ks = NArray(
        label="Ks",
        default=numpy.array([0.0]),
        doc="Slow permittivity coupling strength, the default is no coupling")

    N = NArray(
        dtype=int,
        label="N",
        default=numpy.array([1]),
        doc="The branch of the resting state, default is 1")

    modification = NArray(
        dtype=bool,
        label="modification",
        default=numpy.array([True]),
        doc="When modification is True, then use the modification to stabilise the system for negative values of "
            "dstar. If modification is False, then don't use the modification. The default value is True ")

    state_variable_range = Final(
        label="State variable ranges [lo, hi]",
        default={"x": numpy.array([0.4, 0.6]),
                 "y": numpy.array([-0.1, 0.1]),
                 "z": numpy.array([0.0, 0.1]),
                 "uA": numpy.array([0.0, 0.0]),
                 "uB": numpy.array([0.0, 0.0])},
        doc="Typical bounds on state variables.")

    variables_of_interest = List(
        of=str,
        label="Variables watched by Monitors",
        choices=('x', 'y', 'z'),
        default=('x', 'z'),
        doc="Quantities available to monitor.")

    # state variables names
    state_variables = ('x', 'y', 'z', 'uA', 'uB')

    # number of state variables
    _nvar = 5
    cvar = numpy.array([0], dtype=numpy.int32)

    # If there are derived parameters from the predefined parameters, then initialize them to None
    G = None
    H = None
    L = None
    M = None

    def update_derived_parameters(self):
        r"""
        The equations were adapted from [Saggioetal_2017]
        cf. Eqn. (7), page 17 and page 21

        We parametrize the great arc on the sphere of radius R between the
        points Ain and Aend with the vectors G and H. This great arc is used
        for the offset point of the burster, given by the vector A.

            .. math::
                G &= Ain/\|Ain\| \\
                H &= ((Ain \times Aend) \times Ain)/\|(Ain \times Aend) \times Ain\|

        We also parametrize the great arc on the sphere of radius R between the
        points Bin and Bend with the vectors L and M. This great arc is used
        for the onset point of the burster, given by the vector B.

            .. math::
                L &= Bin/\|Bin\| \\
                M &= ((Bin \times Bend) \times Bin)/\|(Bin \times Bend) \times Bin\|
        """

        Ain = numpy.array([self.mu2_Ain[0], -self.mu1_Ain[0], self.nu_Ain[0]])
        Bin = numpy.array([self.mu2_Bin[0], -self.mu1_Bin[0], self.nu_Bin[0]])
        Aend = numpy.array(
            [self.mu2_Aend[0], -self.mu1_Aend[0], self.nu_Aend[0]])
        Bend = numpy.array(
            [self.mu2_Bend[0], -self.mu1_Bend[0], self.nu_Bend[0]])

        self.G = Ain / numpy.linalg.norm(Ain)
        self.H = numpy.cross(numpy.cross(Ain, Aend), Ain)
        self.H = self.H / numpy.linalg.norm(self.H)

        self.L = Bin / numpy.linalg.norm(Bin)
        self.M = numpy.cross(numpy.cross(Bin, Bend), Bin)
        self.M = self.M / numpy.linalg.norm(self.M)

    def _numpy_dfun(self, state_variables, coupling, local_coupling=0.0):
        x = state_variables[0, :]
        y = state_variables[1, :]
        z = state_variables[2, :]
        uA = state_variables[3, :]
        uB = state_variables[4, :]

        A = self.R * (self.G * numpy.cos(uA) + self.H * numpy.sin(uA))
        B = self.R * (self.L * numpy.cos(uB) + self.M * numpy.sin(uB))

        E = A / (numpy.linalg.norm(A, axis=1)).reshape(-1, 1)
        C = numpy.cross(A,B)
        F = numpy.cross(numpy.cross(A, B), A)
        F = F / (numpy.linalg.norm(F, axis=1)).reshape(-1, 1)

        # Computes the values of mu2,mu1 and nu given the great arc (E,F,R) and the value of the slow variable z
        mu2 = self.R * (numpy.array([E[:, 0]]).T * numpy.cos(z) + numpy.array(
            [F[:, 0]]).T * numpy.sin(z))
        mu1 = -self.R * (numpy.array([E[:, 1]]).T * numpy.cos(z) + numpy.array(
            [F[:, 1]]).T * numpy.sin(z))
        nu = self.R * (numpy.array([E[:, 2]]).T * numpy.cos(z) + numpy.array(
            [F[:, 2]]).T * numpy.sin(z))

        # Computes x_s, which is the solution to x_s^3 - mu2*x_s - mu1 = 0
        if self.N == 1:
            xs = (mu1 / 2.0 + numpy.sqrt(
                mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) + (mu1 / 2.0 - numpy.sqrt(
                mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0)
        elif self.N == 2:
            xs = -1.0 / 2.0 * (1.0 - 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 + numpy.sqrt(
                mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) - 1.0 / 2.0 * (
                1.0 + 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 - numpy.sqrt(mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (
                1.0 / 3.0)
        elif self.N == 3:
            xs = -1.0 / 2.0 * (1.0 + 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 + numpy.sqrt(
                mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) - 1.0 / 2.0 * (
                1.0 - 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 - numpy.sqrt(mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (
                1.0 / 3.0)
        xs = numpy.real(xs)

        # global coupling: To be implemented

        xdot = -y
        ydot = x ** 3 - mu2 * x - mu1 - y * (nu + self.b * x + x ** 2)
        if self.modification:
            zdot = -self.c * (
                numpy.sqrt((x - xs) ** 2 + y ** 2) - self.dstar + 0.1 * (
                    z - 0.5) ** 7 + self.Ks * coupling[0, :])
        else:
            zdot = -self.c * (numpy.sqrt((x - xs) ** 2 + y ** 2) - self.dstar + self.Ks * coupling[0, :])
        uAdot = numpy.full_like(uA, self.cA)
        uBdot = numpy.full_like(uB, self.cB)

        derivative = numpy.array([xdot, ydot, zdot, uAdot, uBdot])
        return derivative

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
        The equations were taken from [Saggioetal_2017]
        cf. Eqns. (4) and (7), page 17 and 21

        The state variables :math:`x` and :math:`y` correspond to the fast subsystem and the
        state variable z corresponds to the slow subsystem. The state
        variables :math:`uA` and :math:`uB` correspond to the transition of the offset and
        onset bifurcations.

            .. math::
                \dot{x} &= -y \\
                \dot{y} &= x^3 - \mu_2 x - \mu_1 - y(\nu + b x + x^2) \\
                \dot{z} &= -c(\sqrt{x-x_{s}^2+y^2} - d^*)\\
                (\dot{u}A) &= cA\\
                (\dot{u}B) &= cB\\

        If the bool modification is True, then the equation for :math:`\dot{z}` will
        been modified to ensure stability for negative :math:`d^*`

            .. math::
                    \dot{z} = -c(\sqrt{x-x_{s}^2+y^2} - d^* + 0.1(z-0.5)^7)

        Where :math:`\mu_1, \mu_2` and :math:`\nu` lie on a great arc of a
        sphere of radius R parametrised by the unit vectors E and F.

            .. math::
                \begin{pmatrix}\mu_2 & -\mu_1 & \nu \end{pmatrix} = R(E \cos z + F \sin z)

        Where the unit vectors E and F are given by:

            .. math::
                E &= A/\|A\| \\
                F &= ((A \times B) \times A)/\|(A \times B) \times A\|

        The vectors A and B transition across a great arc of the same sphere
        of radius R parametrised by G, H and L, M respectively.

            .. math::
                A &= R(G \cos(uA) + H \sin(uA))
                B &= R(L \cos(uB) + M \sin(uB))

        Finally :math:`x_s` is the x-coordinate of the resting state
        (stable equilibrium). This is computed by finding the solution of

            .. math::
                x_s^3 - mu_2*x_s - mu_1 = 0

        And taking the branch which corresponds to the resting state.
        If :math:`x_s` is complex, we take the real part.

        """
        state_variables_ = state_variables.reshape(state_variables.shape[:-1]).T
        coupling_ = coupling.reshape(coupling.shape[:-1]).T
        derivative = _numba_dfun_slowmod(state_variables_, coupling_, self.G[0], self.G[1], self.G[2], self.H[0],
                                         self.H[1], self.H[2], self.L[0], self.L[1], self.L[2], self.M[0], self.M[1],
                                         self.M[2], self.b, self.R, self.c, self.cA, self.cB, self.dstar, self.Ks,
                                         self.modification, self.N)
        return derivative.T[..., numpy.newaxis]


@guvectorize([(float64[:], float64[:], float64[:], float64[:], float64[:], float64[:], float64[:], float64[:],
               float64[:], float64[:], float64[:], float64[:], float64[:], float64[:], float64[:], float64[:],
               float64[:], float64[:], float64[:], float64[:], float64[:], int64[:], int64[:], float64[:])],
             '(n),(m)' + ',()' * 21 + '->(n)', nopython=True)
def _numba_dfun_slowmod(state_variables, coupling, G0, G1, G2, H0, H1, H2, L0, L1, L2, M0, M1, M2, b, R, c, cA, cB,
                        dstar, Ks, modification, N, derivative):
    """Gufunction for the Epileptor Codim 3 model with ultra-slow modulation of classes"""

    x = state_variables[0]
    y = state_variables[1]
    z = state_variables[2]
    uA = state_variables[3]
    uB = state_variables[4]

    A = R[0] * (numpy.array([G0[0], G1[0], G2[0]]) * numpy.cos(uA) + numpy.array([H0[0], H1[0], H2[0]]) * numpy.sin(uA))
    B = R[0] * (numpy.array([L0[0], L1[0], L2[0]]) * numpy.cos(uB) + numpy.array([M0[0], M1[0], M2[0]]) * numpy.sin(uB))

    E = A / (numpy.linalg.norm(A))
    # Numba does not support numpy.cross so we compute the cross-product using the standard formula.
    C = numpy.array([A[1] * B[2] - A[2] * B[1], A[2] * B[0] - A[0] * B[2], A[0] * B[1] - A[1] * B[0]])
    F = numpy.array([C[1] * A[2] - C[2] * A[1], C[2] * A[0] - C[0] * A[2], C[0] * A[1] - C[1] * A[0]])
    F = F / (numpy.linalg.norm(F))

    # Computes the values of mu2,mu1 and nu given the great arc (E,F,R) and the value of the slow variable z
    mu2 = R[0] * (E[0] * numpy.cos(z) + F[0] * numpy.sin(z))
    mu1 = -R[0] * (E[1] * numpy.cos(z) + F[1] * numpy.sin(z))
    nu = R[0] * (E[2] * numpy.cos(z) + F[2] * numpy.sin(z))

    # Computes x_s, which is the solution to x_s^3 - mu2*x_s - mu1 = 0
    if N[0] == 1:
        xs = (mu1 / 2.0 + numpy.sqrt(
            mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) + (mu1 / 2.0 - numpy.sqrt(
            mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0)
    elif N[0] == 2:
        xs = -1.0 / 2.0 * (1.0 - 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 + numpy.sqrt(
            mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) - 1.0 / 2.0 * (
            1.0 + 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 - numpy.sqrt(mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (
            1.0 / 3.0)
    elif N[0] == 3:
        xs = -1.0 / 2.0 * (1.0 + 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 + numpy.sqrt(
            mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (1.0 / 3.0) - 1.0 / 2.0 * (
            1.0 - 1j * 3 ** (1.0 / 2.0)) * (mu1 / 2.0 - numpy.sqrt(mu1 ** 2 / 4.0 - mu2 ** 3 / 27.0 + 0 * 1j)) ** (
            1.0 / 3.0)
    xs = xs.real

    derivative[0] = -y
    derivative[1] = x ** 3 - mu2 * x - mu1 - y * (nu + b[0] * x + x ** 2)
    derivative[2] = -c[0] * (
        numpy.sqrt((x - xs) ** 2 + y ** 2) - dstar[0] + modification[0] * 0.1 * (z - 0.5) ** 7 + Ks[0] * coupling[0])
    derivative[3] = cA[0]
    derivative[4] = cB[0]



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#


import numpy
import scipy.sparse
from tvb.basic.readers import try_get_absolute_path, FileReader
from tvb.datatypes import equations, surfaces
from tvb.basic.neotraits.api import HasTraits, Attr, Float, narray_summary_info


class LocalConnectivity(HasTraits):
    """
    A sparse matrix for representing the local connectivity within the Cortex.
    """
    surface = Attr(field_type=surfaces.CorticalSurface, label="Surface")

    matrix = Attr(field_type=scipy.sparse.spmatrix, required=False)

    equation = Attr(
        field_type=equations.FiniteSupportEquation,
        label="Spatial",
        required=False,
        default=equations.Gaussian())

    cutoff = Float(
        label="Cutoff distance (mm)",
        default=40.0,
        doc="Distance at which to truncate the evaluation in mm.")

    # Temporary obj
    matrix_gdist = None

    def compute(self):
        """
        Compute current Matrix.
        """
        self.log.info("Mapping geodesic distance through the LocalConnectivity.")

        # Start with data being geodesic_distance_matrix, then map it through equation
        # Then replace original data with result...
        self.matrix_gdist.data = self.equation.evaluate(self.matrix_gdist.data)

        # Homogenise spatial discretisation effects across the surface
        nv = self.matrix_gdist.shape[0]
        ind = numpy.arange(nv, dtype=int)
        pos_mask = self.matrix_gdist.data > 0.0
        neg_mask = self.matrix_gdist.data < 0.0
        pos_con = self.matrix_gdist.copy()
        neg_con = self.matrix_gdist.copy()
        pos_con.data[neg_mask] = 0.0
        neg_con.data[pos_mask] = 0.0
        pos_contrib = pos_con.sum(axis=1)
        pos_contrib = numpy.array(pos_contrib).squeeze()
        neg_contrib = neg_con.sum(axis=1)
        neg_contrib = numpy.array(neg_contrib).squeeze()
        pos_mean = pos_contrib.mean()
        neg_mean = neg_contrib.mean()
        if ((pos_mean != 0.0 and any(pos_contrib == 0.0)) or
                (neg_mean != 0.0 and any(neg_contrib == 0.0))):
            msg = "Cortical mesh is too coarse for requested LocalConnectivity."
            self.log.warning(msg)
            bad_verts = ()
            if pos_mean != 0.0:
                bad_verts = bad_verts + numpy.nonzero(pos_contrib == 0.0)
            if neg_mean != 0.0:
                bad_verts = bad_verts + numpy.nonzero(neg_contrib == 0.0)
            self.log.debug("Problem vertices are: %s" % str(bad_verts))
        pos_hf = numpy.zeros(shape=pos_contrib.shape)
        pos_hf[pos_contrib != 0] = pos_mean / pos_contrib[pos_contrib != 0]
        neg_hf = numpy.zeros(shape=neg_contrib.shape)
        neg_hf[neg_contrib != 0] = neg_mean / neg_contrib[neg_contrib != 0]
        pos_hf_diag = scipy.sparse.csc_matrix((pos_hf, (ind, ind)), shape=(nv, nv))
        neg_hf_diag = scipy.sparse.csc_matrix((neg_hf, (ind, ind)), shape=(nv, nv))
        homogenious_conn = (pos_hf_diag * pos_con) + (neg_hf_diag * neg_con)

        # Then replace unhomogenised result with the spatially homogeneous one...
        if not homogenious_conn.has_sorted_indices:
            homogenious_conn.sort_indices()

        self.matrix = homogenious_conn

    @staticmethod
    def from_file(source_file="local_connectivity_16384.mat"):

        result = LocalConnectivity()

        source_full_path = try_get_absolute_path("tvb_data.local_connectivity", source_file)
        reader = FileReader(source_full_path)

        result.matrix = reader.read_array(matlab_data_name="LocalCoupling")
        return result

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance
        of this datatype.
        """
        _, _, v = scipy.sparse.find(self.matrix)
        return narray_summary_info(v, ar_name='matrix-nonzero')

    def compute_sparse_matrix(self):
        """
        NOTE: Before calling this method, the surface field
        should already be set on the local connectivity.

        Computes the sparse matrix for this local connectivity.
        """
        if self.surface is None:
            raise AttributeError('Require surface to compute local connectivity.')

        self.matrix_gdist = surfaces.gdist.local_gdist_matrix(
            self.surface.vertices.astype(numpy.float64),
            self.surface.triangles.astype(numpy.int32),
            max_distance=self.cutoff)

        self.compute()
        # Avoid having a large data-set in memory.
        self.matrix_gdist = None



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
The ProjectionMatrices DataTypes.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""
from enum import Enum

from tvb.basic.readers import try_get_absolute_path, FileReader
from tvb.datatypes import surfaces, sensors
from tvb.basic.neotraits.api import HasTraits, Attr, NArray


class ProjectionsType(Enum):
    EEG = "projEEG"
    MEG = "projMEG"
    SEEG = "projSEEG"


class ProjectionMatrix(HasTraits):
    """
    Base DataType for representing a ProjectionMatrix.
    The projection is between a source of type CorticalSurface and a set of Sensors.
    """

    projection_type = Attr(field_type=str)

    brain_skull = Attr(
        field_type=surfaces.BrainSkull,
        label="Brain Skull", default=None, required=False,
        doc="""Boundary between skull and cortex domains.""")

    skull_skin = Attr(
        field_type=surfaces.SkullSkin,
        label="Skull Skin", default=None, required=False,
        doc="""Boundary between skull and skin domains.""")

    skin_air = Attr(
        field_type=surfaces.SkinAir,
        label="Skin Air", default=None, required=False,
        doc="""Boundary between skin and air domains.""")

    conductances = Attr(
        field_type=dict, label="Domain conductances", required=False,
        default={'air': 0.0, 'skin': 1.0, 'skull': 0.01, 'brain': 1.0},
        doc=""" A dictionary representing the conductances of ... """)

    sources = Attr(
        field_type=surfaces.CorticalSurface,
        label="surface or region", default=None)

    sensors = Attr(
        field_type=sensors.Sensors,
        label="Sensors", default=None, required=False,
        doc=""" A set of sensors to compute projection matrix for them. """)

    projection_data = NArray(label="Projection Matrix Data", default=None, required=True)

    @property
    def shape(self):
        return self.projection_data.shape

    @classmethod
    def from_file(cls, source_file, matlab_data_name=None, is_brainstorm=False):

        proj = cls()

        source_full_path = try_get_absolute_path("tvb_data.projectionMatrix", source_file)
        reader = FileReader(source_full_path)
        if is_brainstorm:
            proj.projection_data = reader.read_gain_from_brainstorm()
        else:
            proj.projection_data = reader.read_array(matlab_data_name=matlab_data_name)
        return proj


class ProjectionSurfaceEEG(ProjectionMatrix):
    """
    Specific projection, from a CorticalSurface to EEG sensors.
    """

    projection_type = Attr(field_type=str, default=ProjectionsType.EEG.value)

    sensors = Attr(field_type=sensors.SensorsEEG)

    @classmethod
    def from_file(cls, source_file='projection_eeg_65_surface_16k.npy', matlab_data_name="ProjectionMatrix",
                  is_brainstorm=False):
        return ProjectionMatrix.from_file.__func__(cls, source_file, matlab_data_name, is_brainstorm)


class ProjectionSurfaceMEG(ProjectionMatrix):
    """
    Specific projection, from a CorticalSurface to MEG sensors.
    """

    projection_type = Attr(field_type=str, default=ProjectionsType.MEG.value)

    sensors = Attr(field_type=sensors.SensorsMEG)

    @classmethod
    def from_file(cls, source_file='projection_meg_276_surface_16k.npy', matlab_data_name=None, is_brainstorm=False):
        return ProjectionMatrix.from_file.__func__(cls, source_file, matlab_data_name, is_brainstorm)


class ProjectionSurfaceSEEG(ProjectionMatrix):
    """
    Specific projection, from a CorticalSurface to SEEG sensors.
    """

    projection_type = Attr(field_type=str, default=ProjectionsType.SEEG.value)

    sensors = Attr(field_type=sensors.SensorsInternal)

    @classmethod
    def from_file(cls, source_file='projection_seeg_588_surface_16k.npy', matlab_data_name=None, is_brainstorm=False):
        return ProjectionMatrix.from_file.__func__(cls, source_file, matlab_data_name, is_brainstorm)


def make_proj_matrix(proj_type):
    """
    Build a ProjectionMatrix instance, based on an input type
    :param proj_type: one of the supported subtypes
    :return: Instance of the corresponding projectiion matrix class, or None
    """
    if proj_type == ProjectionsType.EEG.value:
        return ProjectionSurfaceEEG()
    elif proj_type == ProjectionsType.MEG.value:
        return ProjectionSurfaceMEG()
    elif proj_type == ProjectionsType.SEEG.value:
        return ProjectionSurfaceSEEG()
    return None



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Environment related checks or operations are to be defined here.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

import os
import sys
from subprocess import Popen, PIPE
from tvb.basic.config.settings import VersionSettings


class Environment(object):
    PYTHON_FOLDER = "python%s.%s" % sys.version_info[:2]

    def is_framework_present(self):
        """
        :return: True when framework classes are present and can be imported.
        """
        framework_present = True
        try:
            from tvb.config.profile_settings import WebSettingsProfile
        except ImportError:
            framework_present = False

        return framework_present

    @staticmethod
    def is_distribution():
        """
        Return True when TVB_Distribution package, False when used from a GitHub clone, Pypi, Conda or Docker
        """

        try:
            import tvb_bin
        except ImportError:
            # No tvb_bin, it means usage from Pypi or Conda Forge
            return False

        try:
            _proc = Popen(["git", "status"], stdout=PIPE, stderr=PIPE)
            if "On branch " in str(_proc.communicate()):
                # usage from GitHub clone directly
                return False
        except Exception:
            pass

        if os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(tvb_bin.__file__)))),
                        "externals"):
            # usage from GitHub clone without got cmd or inside a Docker container (as a mounted volume)
            return False

        try:
            _proc = Popen(["svnversion", "."], stdout=PIPE, stderr=PIPE)
            version = VersionSettings.parse_svn_version(_proc.communicate()[0])
            if version:
                # usage from SVN (deprecated)
                return False
        except Exception:
            pass

        # We default as usage from TVB_Distribution
        return True

    def is_linux_deployment(self):
        """
        Return True if current run is not development and is running on Linux.
        """
        return self.is_linux() and self.is_distribution()

    def is_mac_deployment(self):
        """
        Return True if current run is not development and is running on Mac OS X
        """
        return self.is_mac() and self.is_distribution()

    def is_windows_deployment(self):
        """
        Return True if current run is not development and is running on Windows.
        """
        return self.is_windows() and self.is_distribution()

    def is_linux(self):
        return not self.is_windows() and not self.is_mac()

    @staticmethod
    def is_mac():
        return sys.platform == 'darwin'

    @staticmethod
    def is_windows():
        return sys.platform.startswith('win')

    def get_library_folder(self, default_mac):
        """
        Return top level library folder. Will be use for setting paths
        """
        if self.is_windows_deployment():
            return os.path.dirname(sys.executable)
        if self.is_mac_deployment():
            return os.path.dirname(default_mac)
        if self.is_linux_deployment():
            return os.path.dirname(sys.executable)

    def setup_tk_tcl_environ(self, root_folder):
        """
        Given a root folder to look in, find the required configuration files for TCL/TK and set the proper
        environmental variables so everything works fine in the distribution package.

        :param root_folder: the top folder from which to start looking for the required configuration files
        """
        tk_folder = self._find_file('tk.tcl', root_folder)
        if tk_folder:
            os.environ['TK_LIBRARY'] = tk_folder

        tcl_folder = self._find_file('init.tcl', root_folder)
        if tcl_folder:
            os.environ['TCL_LIBRARY'] = tcl_folder

    def _find_file(self, target_file, root_folder):
        """
        Search for a file in a folder directory. Return the folder in which the file can be found.

        :param target_file: the name of the file that is searched
        :param root_folder: the top lever folder from which to start searching in all it's subdirectories
        :returns: the name of the folder in which the file can be found
        """
        for root, _, files in os.walk(root_folder):
            for file_n in files:
                if file_n == target_file:
                    return root

    def setup_python_path(self, *paths):
        """
        Set PYTHONPATH
        :param paths: list of absolute folder paths to join.
        """
        os.environ['PYTHONPATH'] = os.pathsep.join(paths)

    def append_to_path(self, *paths):
        """
        Set PATH
        :param paths: list of absolute folder paths to join and add BEFORE the current PATH
        """
        paths = list(paths)
        paths.append(os.environ.get('PATH', ''))
        os.environ['PATH'] = os.pathsep.join(paths)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)

"""
Wilson-Cowan equations based model definition.

"""
import numpy
from .base import Model
from tvb.basic.neotraits.api import NArray, Final, List, Range


class WilsonCowan(Model):
    r"""
    **References**:

    .. [WC_1972] Wilson, H.R. and Cowan, J.D. *Excitatory and inhibitory
        interactions in localized populations of model neurons*, Biophysical
        journal, 12: 1-24, 1972.
    .. [WC_1973] Wilson, H.R. and Cowan, J.D  *A Mathematical Theory of the
        Functional Dynamics of Cortical and Thalamic Nervous Tissue*

    .. [D_2011] Daffertshofer, A. and van Wijk, B. *On the influence of
        amplitude on the connectivity between phases*
        Frontiers in Neuroinformatics, July, 2011

    Used Eqns 11 and 12 from [WC_1972]_ in ``dfun``.  P and Q represent external
    inputs, which when exploring the phase portrait of the local model are set
    to constant values. However in the case of a full network, P and Q are the
    entry point to our long range and local couplings, that is, the  activity
    from all other nodes is the external input to the local population.

    The default parameters are taken from figure 4 of [WC_1972]_, pag. 10

    In [WC_1973]_ they present a model of neural tissue on the pial surface is.
    See Fig. 1 in page 58. The following local couplings (lateral interactions)
    occur given a region i and a region j:

      E_i-> E_j
      E_i-> I_j
      I_i-> I_j
      I_i-> E_j


    +---------------------------+
    |          Table 1          |
    +--------------+------------+
    |                           |
    |  SanzLeonetAl,   2014     |
    +--------------+------------+
    |Parameter     |  Value     |
    +==============+============+
    | k_e, k_i     |    1.00    |
    +--------------+------------+
    | r_e, r_i     |    0.00    |
    +--------------+------------+
    | tau_e, tau_i |    10.0    |
    +--------------+------------+
    | c_1          |    10.0    |
    +--------------+------------+
    | c_2          |    6.0     |
    +--------------+------------+
    | c_3          |    1.0     |
    +--------------+------------+
    | c_4          |    1.0     |
    +--------------+------------+
    | a_e, a_i     |    1.0     |
    +--------------+------------+
    | b_e, b_i     |    0.0     |
    +--------------+------------+
    | theta_e      |    2.0     |
    +--------------+------------+
    | theta_i      |    3.5     |
    +--------------+------------+
    | alpha_e      |    1.2     |
    +--------------+------------+
    | alpha_i      |    2.0     |
    +--------------+------------+
    | P            |    0.5     |
    +--------------+------------+
    | Q            |    0       |
    +--------------+------------+
    | c_e, c_i     |    1.0     |
    +--------------+------------+
    | alpha_e      |    1.2     |
    +--------------+------------+
    | alpha_i      |    2.0     |
    +--------------+------------+
    |                           |
    |  frequency peak at 20  Hz |
    |                           |
    +---------------------------+


    The parameters in Table 1 reproduce Figure A1 in  [D_2011]_
    but set the limit cycle frequency to a sensible value (eg, 20Hz).

    Model bifurcation parameters:
        * :math:`c_1`
        * :math:`P`



    The models (:math:`E`, :math:`I`) phase-plane, including a representation of
    the vector field as well as its nullclines, using default parameters, can be
    seen below:

        .. _phase-plane-WC:
        .. figure :: img/WilsonCowan_01_mode_0_pplane.svg
            :alt: Wilson-Cowan phase plane (E, I)

            The (:math:`E`, :math:`I`) phase-plane for the Wilson-Cowan model.


    The general formulation for the \textit{\textbf{Wilson-Cowan}} model as a
    dynamical unit at a node $k$ in a BNM with $l$ nodes reads:

    .. math::
            \dot{E}_k &= \dfrac{1}{\tau_e} (-E_k  + (k_e - r_e E_k) \mathcal{S}_e (\alpha_e \left( c_{ee} E_k - c_{ei} I_k  + P_k - \theta_e + \mathbf{\Gamma}(E_k, E_j, u_{kj}) + W_{\zeta}\cdot E_j + W_{\zeta}\cdot I_j\right) ))\\
            \dot{I}_k &= \dfrac{1}{\tau_i} (-I_k  + (k_i - r_i I_k) \mathcal{S}_i (\alpha_i \left( c_{ie} E_k - c_{ee} I_k  + Q_k - \theta_i + \mathbf{\Gamma}(E_k, E_j, u_{kj}) + W_{\zeta}\cdot E_j + W_{\zeta}\cdot I_j\right) ))

    """

    # Define traited attributes for this model, these represent possible kwargs.
    c_ee = NArray(
        label=":math:`c_{ee}`",
        default=numpy.array([12.0]),
        domain=Range(lo=11.0, hi=16.0, step=0.01),
        doc="""Excitatory to excitatory  coupling coefficient""")

    c_ei = NArray(
        label=":math:`c_{ei}`",
        default=numpy.array([4.0]),
        domain=Range(lo=2.0, hi=15.0, step=0.01),
        doc="""Inhibitory to excitatory coupling coefficient""")

    c_ie = NArray(
        label=":math:`c_{ie}`",
        default=numpy.array([13.0]),
        domain=Range(lo=2.0, hi=22.0, step=0.01),
        doc="""Excitatory to inhibitory coupling coefficient.""")

    c_ii = NArray(
        label=":math:`c_{ii}`",
        default=numpy.array([11.0]),
        domain=Range(lo=2.0, hi=15.0, step=0.01),
        doc="""Inhibitory to inhibitory coupling coefficient.""")

    tau_e = NArray(
        label=r":math:`\tau_e`",
        default=numpy.array([10.0]),
        domain=Range(lo=0.0, hi=150.0, step=0.01),
        doc="""Excitatory population, membrane time-constant [ms]""")

    tau_i = NArray(
        label=r":math:`\tau_i`",
        default=numpy.array([10.0]),
        domain=Range(lo=0.0, hi=150.0, step=0.01),
        doc="""Inhibitory population, membrane time-constant [ms]""")

    a_e = NArray(
        label=":math:`a_e`",
        default=numpy.array([1.2]),
        domain=Range(lo=0.0, hi=1.4, step=0.01),
        doc="""The slope parameter for the excitatory response function""")

    b_e = NArray(
        label=":math:`b_e`",
        default=numpy.array([2.8]),
        domain=Range(lo=1.4, hi=6.0, step=0.01),
        doc="""Position of the maximum slope of the excitatory sigmoid function""")

    c_e = NArray(
        label=":math:`c_e`",
        default=numpy.array([1.0]),
        domain=Range(lo=1.0, hi=20.0, step=1.0),
        doc="""The amplitude parameter for the excitatory response function""")

    theta_e = NArray(
        label=r":math:`\theta_e`",
        default=numpy.array([0.0]),
        domain=Range(lo=0.0, hi=60., step=0.01),
        doc="""Excitatory threshold""")

    a_i = NArray(
        label=":math:`a_i`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=2.0, step=0.01),
        doc="""The slope parameter for the inhibitory response function""")

    b_i = NArray(
        label=r":math:`b_i`",
        default=numpy.array([4.0]),
        domain=Range(lo=2.0, hi=6.0, step=0.01),
        doc="""Position of the maximum slope of a sigmoid function [in
        threshold units]""")

    theta_i = NArray(
        label=r":math:`\theta_i`",
        default=numpy.array([0.0]),
        domain=Range(lo=0.0, hi=60.0, step=0.01),
        doc="""Inhibitory threshold""")

    c_i = NArray(
        label=":math:`c_i`",
        default=numpy.array([1.0]),
        domain=Range(lo=1.0, hi=20.0, step=1.0),
        doc="""The amplitude parameter for the inhibitory response function""")

    r_e = NArray(
        label=":math:`r_e`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.5, hi=2.0, step=0.01),
        doc="""Excitatory refractory period""")

    r_i = NArray(
        label=":math:`r_i`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.5, hi=2.0, step=0.01),
        doc="""Inhibitory refractory period""")

    k_e = NArray(
        label=":math:`k_e`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.5, hi=2.0, step=0.01),
        doc="""Maximum value of the excitatory response function""")

    k_i = NArray(
        label=":math:`k_i`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=2.0, step=0.01),
        doc="""Maximum value of the inhibitory response function""")

    P = NArray(
        label=":math:`P`",
        default=numpy.array([0.0]),
        domain=Range(lo=0.0, hi=20.0, step=0.01),
        doc="""External stimulus to the excitatory population.
        Constant intensity.Entry point for coupling.""")

    Q = NArray(
        label=":math:`Q`",
        default=numpy.array([0.0]),
        domain=Range(lo=0.0, hi=20.0, step=0.01),
        doc="""External stimulus to the inhibitory population.
        Constant intensity.Entry point for coupling.""")

    alpha_e = NArray(
        label=r":math:`\alpha_e`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=20.0, step=0.01),
        doc="""External stimulus to the excitatory population.
        Constant intensity.Entry point for coupling.""")

    alpha_i = NArray(
        label=r":math:`\alpha_i`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=20.0, step=0.01),
        doc="""External stimulus to the inhibitory population.
        Constant intensity.Entry point for coupling.""")

    # Used for phase-plane axis ranges and to bound random initial() conditions.
    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"E": numpy.array([0.0, 1.0]),
                 "I": numpy.array([0.0, 1.0])},
        doc="""The values for each state-variable should be set to encompass
        the expected dynamic range of that state-variable for the current
        parameters, it is used as a mechanism for bounding random inital
        conditions when the simulation isn't started from an explicit history,
        it is also provides the default range of phase-plane plots.""")

    variables_of_interest = List(
        of=str,
        label="Variables watched by Monitors",
        choices=("E", "I", "E + I", "E - I"),
        default=("E",),
        doc="""This represents the default state-variables of this Model to be
               monitored. It can be overridden for each Monitor if desired. The
               corresponding state-variable indices for this model are :math:`E = 0`
               and :math:`I = 1`.""")

    state_variables = 'E I'.split()
    _nvar = 2
    cvar = numpy.array([0, 1], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""

        .. math::
            \tau \dot{x}(t) &= -z(t) + \phi(z(t)) \\
            \phi(x) &= \frac{c}{1-exp(-a (x-b))}

        """

        E = state_variables[0, :]
        I = state_variables[1, :]
        derivative = numpy.empty_like(state_variables)

        # long-range coupling
        c_0 = coupling[0, :]

        # short-range (local) coupling
        lc_0 = local_coupling * E
        lc_1 = local_coupling * I

        x_e = self.alpha_e * (self.c_ee * E - self.c_ei * I + self.P  - self.theta_e +  c_0 + lc_0 + lc_1)
        x_i = self.alpha_i * (self.c_ie * E - self.c_ii * I + self.Q  - self.theta_i + lc_0 + lc_1)

        s_e = self.c_e / (1.0 + numpy.exp(-self.a_e * (x_e - self.b_e)))
        s_i = self.c_i / (1.0 + numpy.exp(-self.a_i * (x_i - self.b_i)))

        derivative[0] = (-E + (self.k_e - self.r_e * E) * s_e) / self.tau_e
        derivative[1] = (-I + (self.k_i - self.r_i * I) * s_i) / self.tau_i

        return derivative
